; 
; This program installs a temporary handler for the keyboard
; hardware interrupt and loops, displaying a running count of
; the calls to the interrupt, until the user presses the
; Escape key.
;
; This program should run OK under Windows 9x, but because
; Windows' emulation of a standard BIOS is not exact, the
; interrupt will not be called for the shift or toggle keys.
; This program has not been tested under Windows 2000/XP, or
; under DOS.
;       

      .model small, c
      .386
      .stack

      ; DOS handles for standard input/output devices.
      HSTDIN   EQU 0
      HSTDOUT  EQU 1

      w2dec    PROTO C :WORD,:WORD
      w2hex    PROTO C :WORD,:WORD
      stdin    PROTO C :WORD,:WORD
      stdout   PROTO C :WORD
      szlen    PROTO C :WORD
      locate   PROTO c :BYTE,:BYTE

      ; This is the MASM32 reparg macro with *no* changes.
      ; This macro replaces quoted text with a DATA section
      ; OFFSET and returns it in ADDR "name" format.
      reparg MACRO arg
         LOCAL nustr
         quot SUBSTR <arg>,1,1
         IFIDN quot,<">          ;; if 1st char = "
         .data
            nustr db arg,0    ;; write arg to .DATA section
         .code
         EXITM <ADDR nustr>   ;; append name to ADDR operator
         ELSE
            EXITM <arg>          ;; else return arg
         ENDIF
      ENDM

      ; This is a VARARG version of the MASM32 print macro.
      print MACRO varname:VARARG
         FOR arg,<varname>
            invoke stdout,reparg(arg)
         ENDM
      ENDM

.data
      inbuf       db 130 dup(0)
      outbuf      db 30 dup(0)
      crlf        db 13,10,0
      NL          dw offset crlf
.code
.startup

      ; Place the variables that the handler needs
      ; to access in the code segment. An interrupt
      ; handler can always depend on the value in
      ; CS, but not the values in the other segment
      ; registers.
      jmp   @F
      prevHandler dd 0
      intCount    dw 0
   @@:
      ; Get and save the address of the previous
      ; Interrupt 09h handler.
      mov   ax,3509h
      int   21h
      mov   word ptr prevHandler,bx
      mov   word ptr prevHandler[2],es

      ; Set the Interrupt 09h vector-table entry
      ; to the address of the temporary handler.
      push  ds
      mov   ax,2509h
      push  cs
      pop   ds
      mov   dx,offset TempHandler
      int   21h
      pop   ds

      ; Display the interrupt count until the
      ; user presses Escape.
      call cls
      print  NL,"Press Esc to exit,", \
         "any other key to update the interrupt count"
   @@:
      invoke locate,3,0
      invoke w2dec,intCount,addr outbuf
      print "intCount: ",addr outbuf
      call waitkey
      cmp   al,27
      jne   @B

      ; Restore the Interrupt 09h vector-table
      ; entry to its original state.
      push  ds
      lds   dx,prevHandler
      mov   ax,2509h
      int   21h
      pop   ds

      print NL,NL,"Press any normal key to exit...",NL
      call waitkey
.exit

; 
; This is the temporary Interrupt 9 handler. It will be
; called at least once each time a key is pressed or
; released, and more than once for most of the extended
; keys.
; 
TempHandler:
      ; Increment the interrupt count.
      inc   cs:intCount

      ; Chain to the previous handler with a far jump.
      ; This chaining is necessary to allow the other
      ; handlers in the chain to do whatever they need
      ; to do, the most important being the BIOS
      ; handler.       
      jmp   DWORD PTR cs:prevHandler

; 
; This proc converts a word to to a decimal string
; and stores the result in the specified buffer.
; 
w2dec proc uses ax bx cx dx di src:WORD,npdest:WORD
      mov   ax,src
      mov   di,npdest
      ; Test for negative value.
      test  ax,ax
      jns   @F
      ; Make positive if negative.
      neg   ax   
      ; Store sign in buffer.
      mov   byte ptr[di],'-'
      inc   di
   @@:
      ; Init divisor.
      mov   bx,10
      ; Zero digit counter.
      mov   cx,0
   divLoop:
      ; Zero MSW of dividend.
      xor   dx,dx      
      ; Divide value by 10.
      div   bx
      ; Push remainder.
      push  dx
      ; Increment digit count.
      inc   cx
      ; Repeat if quotient > 0.
      or    ax,ax
      jnz   divLoop
   storeLoop:
      ; Pop digit off stack.
      pop   ax
      ; Convert to ascii.
      add   al,'0'
      ; Store in buffer.
      mov   [di],al
      inc   di
      ; Repeat CX times.
      loop  storeLoop
      ; Append trailing null.
      mov   byte ptr[di],0
      ret
w2dec endp

; 
; This proc converts a word to to a hex string and
; stores the result in the specified buffer.
; 
   .data
      hexChars db "0123456789ABCDEF"
   .code
w2hex proc uses ax bx dx di src:WORD,npdest:WORD
      mov   dx,src
      mov   di,npdest
      ; Set DI to end of buffer.
      add   di,3
      ; Append trailing null.
      mov   byte ptr[di+1],0
   @@:
      ; Copy current value into BX.
      mov   bx,dx
      ; Mask off all but lower nibble.
      and   bx,0Fh
      ; Load corresponding hex character into AL.
      mov   al,hexChars[bx]
      ; And store it in buffer.
      mov   [di],al
      ; Shift next nibble into position.
      shr   dx,4
      ; Adjust DI for next character.
      dec   di
      ; Repeat for all 4 characters.
      cmp   di,npdest
      jnb   @B
      ret
w2hex endp

; 
; This proc calls the Read File or Device function to
; read up to <blen> bytes from STDIN and store them in
; the specified buffer.
; 
stdin proc uses bx cx dx npdest:WORD,blen:WORD
      mov   bx,HSTDIN
      mov   cx,blen
      mov   dx,npdest
      mov   ah,3fh
      int   21h      
      ; Replace CR with null.
      mov   bx,npdest
   @@:
      cmp   byte ptr[bx],0
      je    @F
      cmp   byte ptr[bx],13
      je    @F
      inc   bx
      jmp   @B
   @@:
      mov   byte ptr[bx],0
      ret
stdin endp

; 
; This proc calls the Write File or Device function to
; output the specified string to STDOUT.
; 
stdout proc uses ax bx cx dx npsz:WORD   
      invoke szlen,npsz
      mov   cx,ax
      mov   bx,HSTDOUT
      mov   dx,npsz      
      mov   ah,40h
      int   21h      
      ret
stdout endp

; 
; This proc returns the length of the specified
; null-terminated string in AX.
; 
szlen proc uses bx npsz:WORD
      mov   bx,npsz
   runLoop:
      ; Compare bytes until find null.
      cmp   byte ptr[bx],0
      jz    @F
      inc   bx
      jmp   runLoop
   @@:
      ; Calc length and return in AX.
      mov   ax,bx
      sub   ax,npsz
      ret
szlen endp

; 
; This proc calls the Flush Buffer, Read Keyboard Without
; Echo function and returns the character read in AL.
; 
waitkey proc
      mov   ax,0c08h
      int   21h
      ret
waitkey endp

; 
; This proc calls the BIOS Set Cursor Position function
; to set the cursor position for display page 0 to
; <row>,<col> (upper left corner is 0,0).
; 
locate proc uses ax bx dx row:BYTE,col:BYTE
      mov   ah,2
      mov   bh,0
      mov   dh,row
      mov   dl,col
      int   10h
      ret
locate endp

; 
; This proc clears the screen by calling the BIOS Scroll
; Current Page Up function with a row count of zero, and
; sets the cursor position to the upper left corner of
; the screen. An 80x25 text mode screen is assumed.
; 
cls proc uses ax bx cx dx
      mov   ax,600h
      mov   bh,7
      mov   ch,0
      mov   cl,0      
      mov   dh,24
      mov   dl,79
      int   10h
      invoke locate,0,0
      ret
cls endp

end
